/*Copyright (c) 2008 Jose Antonio Fernandes Salvador (AZTI-Tecnalia).
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

import java.io.*;

import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Utils;
import weka.core.converters.*;
import weka.attributeSelection.*;
import weka.classifiers.*;
import weka.classifiers.bayes.*;
import weka.classifiers.bayes.net.*;
import weka.classifiers.evaluation.NominalPrediction;
import weka.classifiers.evaluation.Prediction;
import weka.classifiers.meta.*;
import weka.core.Instances;
import weka.filters.Filter;
import java.util.*; 

/*
 * For use this class, the class variables have to be at the end 
 * and classindex has to specify the first one
 * */

/*
 * FALTA PROGRAMAR:
 * - Perform prediction
 * - Perform especified class discretization
 * 
 * */

public class MainProgram {
	  static int m_numClasses =3;
	  static int iMinInstancesPerClass =0; //6 ARI  // 5 hake Avoids evaluating extremely umbalance databases that are useless and increase CPU time
	  static String sLearningFile="ariRaw.arff";
	  static String sTopredictFile="";
	  static private int NumDimensions =4;
	  static private int m_NumFolds =5; // 10x10cv
	  static private int m_NumCV = 10; // 10x10cv
	  static private int m_NumBootstrap =100;
	  static private int m_NumAtributes;
	  static private int m_ClassIndex;
	  static private double m_bootStrap_percentSplit = 0.632;
	  static private FastVector m_PredictionsAll;
	  static public NominalPrediction [] m_Predictions;
	  static private ArrayList listCutPoints;
	  static private double m_Fayyad_entropy =0;
	  static boolean vflag = true;
	  static boolean flagConsensualCFS = false;
	  static boolean flagModelBuilding = false;
	  static boolean flagModelValidation = false;
	  static boolean flagPerformPrediction = false;
	  static boolean flagFilterClassDiscretization = false;
	  static boolean flagWrapperAccuracy = false;
	  static boolean flagPerformPredicition = false;
	  static boolean flagSetClassDiscretization=false;
	  static String sExternalCutPoints = "";
  	  static String sFilter="";
	  static double [] m_CutPoints = null;
	  static int iNumCutPoints =0;
	  static ArrayList[] listPredictorsDiscretization; // For returning the list os predictors discretization
	  static ArrayList m_CutPointCombinations;
	  static weka.core.Instances dataSetToPredict = null;
	    
	  
	public MainProgram() {
		// TODO Apndice de constructor generado automticamente
	}

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		// TODO Apndice de mtodo generado automticamente
		try {
			System.out.println("Starting...");
			ParseArguments(args);
			
			File fCargar;
			 CSVLoader csvRead = new CSVLoader();
			 
			 fCargar = new File(sLearningFile);
			 csvRead.setFile(fCargar);
			 weka.core.Instances dataSet = csvRead.getDataSet();
						
		    m_NumAtributes = dataSet.numAttributes();
		    m_ClassIndex =m_NumAtributes-1;
		    dataSet.setClassIndex(m_ClassIndex);
		    if (iMinInstancesPerClass==0)iMinInstancesPerClass = (int)(Math.sqrt(dataSet.numInstances()));
		    
		   if (flagPerformPrediction) {
		    	 fCargar = new File(sTopredictFile);
				 csvRead.setFile(fCargar);
				 dataSetToPredict = csvRead.getDataSet();
		    }
		    
		    if ((sExternalCutPoints!="")&&(dataSet.classAttribute().isNumeric())) {
		    	  String[] result = sExternalCutPoints.split(",");
		    	  ArrayList arrCuts = new ArrayList();
		    	  for (int x=0; x<result.length; x++) arrCuts.add(new Double(result[x]));
		    	  
		    	  TrainTestPair trainTest = DiscretizeFeature(new TrainTestPair(dataSet,null),arrCuts,m_ClassIndex);
		    	  dataSet = trainTest.insTrain;
		    	  dataSet.setClassIndex(m_ClassIndex);
		    }
		    
		    if ((dataSet.classAttribute().isNumeric())&&(flagModelBuilding||flagModelValidation||flagPerformPrediction)) {
				listCutPoints = new ArrayList(); // Arraylis of CutPointSet
				Instances datasetSorted = new Instances(dataSet);
				datasetSorted.sort(m_ClassIndex);
				 m_CutPoints = cutPointsForSubset(datasetSorted,m_ClassIndex); 
				 m_CutPointCombinations= new ArrayList(); // ArrayList of CutPointSetPredictor
				 ValidCutPointCombinations(); 
		    }	 
			 
			 if (dataSet.classAttribute().isNominal()) {
				 m_numClasses = dataSet.numClasses();
			
				 if (flagConsensualCFS) {
					 if (vflag) System.out.println("Starting Consensual CFS of "+ sLearningFile);
					 ConsensualCFS(new TrainTestPair(dataSet,null),new ArrayList(), true);  // Si le llamo directamente, sera la discretizacin del experto
					 if (vflag) System.out.println("Consensual CFS of "+ sLearningFile+" finished");
				 }
			}
			 
			 if (flagModelBuilding||flagPerformPrediction) {
				 if (vflag) System.out.println("Starting Model Building of "+ sLearningFile);
				 InnerLoopModelBuilding(new TrainTestPair(dataSet,null),true);
				 if (vflag) System.out.println("Model Building of "+ sLearningFile+" finished");
			 }
			 if (flagModelValidation){
				 if (vflag) System.out.println("Starting Model Validation of "+ sLearningFile);
				 OuterLoop(dataSet);
				 if (vflag) System.out.println("Model validation of "+ sLearningFile+" finished");
			 }
					 
			
		} catch (Exception e) {
			e.printStackTrace();
			e.getCause();
		}
		
		System.out.println("Finished.");
	  
	}
	
	
	 private static Classifier getTAN(){
     Classifier cls = new BayesNet();
     String[] options = new String[11];
     options[0] = "-D";
     options[1] = "-Q";      
     options[2]="weka.classifiers.bayes.net.search.local.TAN";
     options[3]="--";
     options[4]="-S";
     options[5]="-BAYES";
     options[6]="-E";
     options[7]="weka.classifiers.bayes.net.estimate.SimpleEstimator";
     options[8]="--";
     options[9]="-A";
     options[10]="0.5";           
     try{
         cls.setOptions(options);
     }
     catch (Exception ex){
         System.out.println("Exception on seting options: " + ex.getMessage());
     }    
     return cls;   
 }
	 
	 
	 private static Classifier getNaiveBayes(){
     Classifier cls = new NaiveBayes();

     return cls;   
 }
	 
 
	 static public double[] calculateAccuracyFold(double [][] predictions, int iNumPredictions){
		 double [] PctCorrectAcc = new double[NumDimensions];
		 
		 // Initialization to zeros
		 for (int k = 0; k < NumDimensions; k++) 
				 PctCorrectAcc[k]=0;
		 
		 // Count corrects
		 	 NominalPrediction [] arrPredictions = new NominalPrediction[NumDimensions];
			 
			 for (int i = 0; i < iNumPredictions; i++) {
				 arrPredictions = (NominalPrediction [])m_PredictionsAll.elementAt(i);
				
				 for (int j = 0; j < NumDimensions; j++) 
					 
					 if (arrPredictions[j].actual() ==arrPredictions[j].predicted())
						  PctCorrectAcc[j]++;
			 }
		 
		 // Compute rate (Accuracy for each dimension and fold)
		 for (int k = 0; k < NumDimensions; k++) 
				 PctCorrectAcc[k]=PctCorrectAcc[k]*100/iNumPredictions;
	
		 return PctCorrectAcc;
	 }

	 static public double standardDeviation(double [] arrSamples, int iNumSamples) {
		 double std =0;
		 double std1 =0;
		 double std2 =0;
		 
		 for (int i = 0; i < iNumSamples; i++) {
			 std1+=arrSamples[i]*arrSamples[i];
			 std2+=arrSamples[i];
		 }
		 std1 = std1 * iNumSamples;
		 std2 = std2 * std2;
		 
		 double denominator = iNumSamples * (iNumSamples-1);
		 std = Math.sqrt((std1-std2)/denominator);
		 
		 return std;
	 }
	 
	 static protected double[] cutPointsForSubset(Instances instances, int attIndex) { 
		 int numInstances = instances.numInstances();
		 double[] temp = new double[numInstances];
		 double currentCutPoint;
		 iNumCutPoints=0;
		 
		 for (int iActualIns =0; iActualIns < numInstances-1; iActualIns++) {
			 if (instances.instance(iActualIns).value(attIndex) < instances.instance(iActualIns + 1).value(attIndex)) {
					currentCutPoint = (instances.instance(iActualIns).value(attIndex) + 
					  instances.instance(iActualIns + 1).value(attIndex)) / 2.0;
					
					try {
						temp[iNumCutPoints] = Math.rint(currentCutPoint*100)/100;
					} catch (Exception e) {
						//System.out.println("Error in Boot trying to find cutpoints: "+iIteration);	
						e.printStackTrace();
					}  
					iNumCutPoints ++;
			 }
		 }
		 
		 double[] arrCutPoints = new double[iNumCutPoints];
		 
		 for (int k=0; k<iNumCutPoints; k++) arrCutPoints[k] =temp[k];
		 
		 return arrCutPoints;
	 }
	 
	private static void OuterLoop(Instances dataSet){
		 try{
			 String sFichero = "OuterLoop_"+sLearningFile+sFilter+"_NaiveBayes_NumClasses"+m_numClasses+"_MinInstClass"+iMinInstancesPerClass+"_"+m_NumCV+"CV"+m_NumFolds+"_boots"+m_NumBootstrap+".csv";
			 File fichero = new File(sFichero);
			 if (fichero.exists()) fichero.delete();
			 BufferedWriter bw = new BufferedWriter(new FileWriter(sFichero));
			 String sTP = "";
			 for (int i =1;i<=m_numClasses;i++)	sTP=sTP+",TP"+i;
			 
			 bw.write("Accuracy"+sTP+",MeanTP,BrierScore\n");
			 				 
				double[] PctCorrect= new double[m_NumFolds];
				double[] meanTPs= new double[m_numClasses];
			 	double[] meanPctCorrectCVs = new double[m_NumCV];
			 	double[] meanBrierScoreCVs = new double[m_NumCV];
			 	
				double meanPctCorrect=0;
				double meanBrierScore =0;
				
			 	double meanPctCorrectA =0;
			 	double meanBrierScoreA =0;
			 	
			 	EvaluationBrier evalC;
			 	String sResults ="";
			 	double meanTPsFold =0;
			 				 	
			 	for (int numCV = 0; numCV < m_NumCV; numCV++) {
			 		System.out.println("Starting repeat "+numCV+ " cv: "+new Date());
//			 		Make the folds
			 		 dataSet.setClassIndex(dataSet.numAttributes()-1);
					 
					 Instances runInstances = new Instances(dataSet);
					 if (runInstances.classAttribute().isNominal()) {
					   // runInstances.stratify(m_NumFolds);
					 }
			 		 Random random = new Random(numCV);
					 runInstances.randomize(random);
					
					for (int fold=0;fold<m_NumFolds;fold++){
						 System.out.println("Starting repeat:"+numCV+" fold "+fold+ " cv: "+ new Date());
						//bw.write("Results class " + dataSet.classAttribute() + ":\n");
							 
						Instances train = runInstances.trainCV(m_NumFolds, fold);
						Instances test = runInstances.testCV(m_NumFolds, fold);
						train.setClassIndex(train.numAttributes()-1);
						test.setClassIndex(test.numAttributes()-1);
						TrainTestPair acTrainTest = InnerLoopModelBuilding(new TrainTestPair(train,test),false);
						evalC = ModelValidation(acTrainTest.insTrain, acTrainTest.insTest);
						meanTPsFold =0;	
						meanPctCorrect += evalC.pctCorrect();
						meanBrierScore += evalC.BrierScore();
						sResults ="";
						for (int i =0;i<m_numClasses;i++) {
						  	meanTPs[i]+=evalC.truePositiveRate(i);
						   	//sResults= sResults +evalC.truePositiveRate(i)+",";
						   	meanTPsFold +=evalC.truePositiveRate(i);
						   	sResults = sResults + "," +evalC.truePositiveRate(i);
						}
						bw.write((Math.rint(evalC.pctCorrect()*100)/100)+sResults+","+(Math.rint(meanTPsFold*100)/100)/m_numClasses+","+evalC.BrierScore()+"\n");
					}
					 	
					 	meanPctCorrectCVs[numCV] = meanPctCorrect/m_NumFolds;
					 	meanBrierScoreCVs[numCV] = meanBrierScore/m_NumFolds;
					 	meanPctCorrectA += meanPctCorrect/m_NumFolds;
					 	meanBrierScoreA += meanBrierScore/m_NumFolds;
					 	meanPctCorrect =0;
					 	meanBrierScore =0;
			 	}
		 	 	AttributeSelectionMerit attsel2= new AttributeSelectionMerit();
				CfsSubsetEval eval2 = new CfsSubsetEval();
				BestFirstMulti search2 = new BestFirstMulti();
				attsel2.setEvaluator(eval2);
				attsel2.setSearch(search2);
				attsel2.SelectAttributes(dataSet);
				double dMerit2 = attsel2.m_bestMerit;
				int [] indicesFull2 = attsel2.selectedAttributes();
				String sSelectedAtt = "";
				for (int iAtt=0; iAtt<indicesFull2.length-1;iAtt++) {
					sSelectedAtt = sSelectedAtt+ dataSet.attribute(indicesFull2[iAtt]).name()+ ";";
				}
			 	
		 		// for two decimals	(Math.rint(meanAllDimensions*100)/100)
		 		meanPctCorrectA = (Math.rint((meanPctCorrectA / m_NumCV)*100)/100);
		 		meanBrierScoreA = (Math.rint((meanBrierScoreA / m_NumCV)*10000)/10000);
		 		double meanTPglobal =0; sResults ="";
		 		for (int i =0;i<m_numClasses;i++) {
			    	meanTPs[i]=(Math.rint((meanTPs[i]/(m_NumFolds*m_NumCV))*1000)/1000);
			    	meanTPglobal +=meanTPs[i];
			    	sResults = sResults +"; TP"+i+":"+meanTPs[i];
		 		}
			 				 	
		 		meanTPglobal = (Math.rint((meanTPglobal/m_numClasses)*1000)/1000); 
			 	double sd =(Math.rint((standardDeviation(meanPctCorrectCVs,m_NumCV))*1000)/1000);
			 	double sdBrier =(Math.rint((standardDeviation(meanBrierScoreCVs,m_NumCV))*1000)/1000);
				
			 	
			 	System.out.println("Acc cv: "+meanPctCorrectA+" ("+sd+"); Brier score cv: "+(Math.rint(meanBrierScoreA*1000)/1000)+" ("+sdBrier+"); "+sResults+ "; MeanTPs:"+meanTPglobal);
			    bw.write("Acc cv: "+meanPctCorrectA+" ("+sd+"); Brier score cv: "+(Math.rint(meanBrierScoreA*1000)/1000)+" ("+sdBrier+"); "+sResults+ "; MeanTPs:"+meanTPglobal);
			    for (int i =0;i<m_numClasses;i++) {
			    	meanTPs[i]=0;
			    }
			 bw.close();
		 } catch (Exception ex){
			 ex.printStackTrace();
	    }   
	 }
	 
	 private static TrainTestPair InnerLoopModelBuilding(TrainTestPair acTrainTestPair, boolean bOutputToFile){
		 CutPointSet cutBestCutpoint= null; 
		 				
		 double dMaxPerformance =0;
		 TrainTestPair rtnTrainTestPair;
		 if (acTrainTestPair.insTrain.classAttribute().isNumeric()) {
			 try {
				 listCutPoints= new ArrayList();
	//			 acTrainTestPair.insTrain.sort(acTrainTestPair.insTrain.numAttributes()-1);
				 String sFichero;
				
				 sFichero= "ClassDiscretization_"+sLearningFile+sFilter+"_NaiveBayes_NumClasses"+m_numClasses+"_MinInstClass"+iMinInstancesPerClass+"_boots"+m_NumBootstrap+".csv"; // Put an index for each cv
				 
				 File fichero = new File("temp.csv");
				 BufferedWriter bw= new BufferedWriter(new FileWriter("temp.csv"));
				 if (bOutputToFile){
					 bw.close();
					 fichero = new File(sFichero);
					 if (fichero.exists()) fichero.delete();
					 bw= new BufferedWriter(new FileWriter(sFichero));
						
				 }
				 
				 
				 if (flagFilterClassDiscretization){ // Filter discretization
					 String sCutPoints="", sIns ="",sFSS1="",dFSS1Temp="";
					 int iFSS1 =0, iNumFeatures=0;
					 double dFSS1merit=0,dMeanEntropy=0;
					 if (bOutputToFile){
						 for (int i =1;i<=m_numClasses;i++){
							 if (i<m_numClasses) sCutPoints=sCutPoints+"CutP"+i+",";
							 sIns =sIns+"#Ins"+i+",";;
						 }
						 if (bOutputToFile) {							
							 String sHeaders =sCutPoints+"TotalEntropyAfterCFS,MeanEntropyAfterCFS,FSSatributes1,#Selected1,FSSmerit1\n";
							 bw.write(sHeaders);
						 }
					 }
					 ClassDiscretizationFilter(acTrainTestPair,0,false);
					 
					 Iterator iterCutPoints = listCutPoints.iterator();
					 CutPointSet cutpoint;
					 while (iterCutPoints.hasNext()) {
						 sCutPoints =""; 
						 cutpoint = (CutPointSet)iterCutPoints.next();
						 
						 ArrayList arrTempFSSlist =cutpoint.listFSS;
						 if (arrTempFSSlist.size()>0) {
							 sFSS1 =((SubsetConsensual)arrTempFSSlist.get(0)).getSubsetNames();
							 iFSS1 =((SubsetConsensual)arrTempFSSlist.get(0)).getSelectedCounts();
							 dFSS1merit =((SubsetConsensual)arrTempFSSlist.get(0)).getMerit();
							 dFSS1Temp =sFSS1;
							 iNumFeatures=0;
							 while (dFSS1Temp.contains(";")){
								 dFSS1Temp =dFSS1Temp.replaceFirst(";","");
								 iNumFeatures++;
							 }
							 dMeanEntropy = cutpoint.dFayyadEntropy/iNumFeatures;
							 
						 }
						 
						 for (int i=0;i<m_numClasses;i++){ if (i<m_numClasses-1) sCutPoints=sCutPoints+cutpoint.listPoints.get(i)+",";
						 	sIns=sIns+(Math.rint(cutpoint.arrClassPriors[i]*10)/10)+",";
						 }
						 
						 if (bOutputToFile)	bw.write(sCutPoints+cutpoint.dFayyadEntropy+","+dMeanEntropy+","+sFSS1+","+iFSS1+","+dFSS1merit+"\n"); 
						 
						 if (cutBestCutpoint==null) {
								cutBestCutpoint = cutpoint;
								dMaxPerformance = dMeanEntropy; // If bestcutpoint is the one with best MeanTP
							} else {
								if ((dMaxPerformance>dMeanEntropy)||(dMaxPerformance==0)) { // The entropy has to be minimized
										cutBestCutpoint = cutpoint;
										dMaxPerformance = dMeanEntropy; //If bestcutpoint is the one with best MeanTP
								}
							}
					 }
				 } else {
					 int n =  acTrainTestPair.insTrain.numInstances();
					 
					 for (int boots=0;boots<m_NumBootstrap;boots++){
						 System.out.println("Bootstrapp number: "+ boots+", time: "+ new Date());
						 int[] ind = new int[n];
						 Random randomBoot = new Random(boots);
						 
						 for (int i=0;i<n;i++){
							 ind[i] = randomBoot.nextInt(n);
						 }
						 Arrays.sort(ind);
						 weka.core.Instances bootsTrain = new weka.core.Instances(acTrainTestPair.insTrain,0);
						 weka.core.Instances bootsTest = new weka.core.Instances(acTrainTestPair.insTrain,0);
						 int iActualShouldBe = 0;
						 int lastAdded = -1;
						 for (int i=0;i<n;i++){
							 if (ind[i]<=iActualShouldBe){
								 bootsTrain.add(acTrainTestPair.insTrain.instance(ind[i]));
								 if ((i==0)||(ind[i]>ind[i-1]))iActualShouldBe++;
							 } else {
								 while (ind[i] > iActualShouldBe) {
									 bootsTest.add(acTrainTestPair.insTrain.instance(iActualShouldBe));
									 iActualShouldBe++;
									 
								 } 
								 if (ind[i]==iActualShouldBe){
									 bootsTrain.add(acTrainTestPair.insTrain.instance(ind[i]));
									 iActualShouldBe++;
								 }
							 }
							 
							 
						 }
						
						// Class discretization, predictors discretization and CFS
						ClassDiscretization(new TrainTestPair(bootsTrain,bootsTest),boots,bOutputToFile);
						System.gc();
					
					 }
					 
					 // Calculate for each cutpoint performance (trought bootstrap) and save to file
					  String sCutPoints="";String sIns="";String sTP="";String sROC="";String sFmeasure="";String sRecall="";String sPrecision="";
					 if (bOutputToFile){
						 for (int i =1;i<=m_numClasses;i++){
							 if (i<m_numClasses) sCutPoints=sCutPoints+"CutP"+i+",";
							 sIns =sIns+"#Ins"+i+",";sTP=sTP+",TP"+i;sROC=sROC+",ROC"+i;sFmeasure=sFmeasure+",F"+i+"measure"+i;sRecall=sRecall+",Recall"+i;sPrecision=sPrecision+",Prec"+i+"ision"+i;
						 }
						 String sHeaders =sCutPoints+sIns+"DiscretAllVarsTotalEntropy"+sTP+",MeanTP,#Eval,Accuracy,Brier"+sROC+",MeanRoc"+sFmeasure+",MeanFmeasure"+sRecall+",MeanRecall"+sPrecision+",MeanPrecision,RootMeanSquareError,RelativeAbsoluteError,RootMeanPriorSquaredError,RootRelativeSquaredError,KBRelatieveInformation,KBInformation,KBMeanInformation,Kappa,FSSatributes1,#Selected1,FSSmerit1,FSSatributes2,#Selected2,FSSmerit2\n";
						 
						 bw.write(sHeaders);
					 }
					 Iterator iterCutPoints = listCutPoints.iterator();
					 CutPointSet cutpoint;
					 double dMeanTP=0,dMeanROC=0,dMeanFmeasure=0,dMeanRecall=0,dMeanPrecision=0;
					 while (iterCutPoints.hasNext()) {
						 	cutpoint= (CutPointSet)iterCutPoints.next();
						 	cutpoint.dAccuracyE0 = cutpoint.dAccuracyE0 *0.632/cutpoint.iNumPerformances;
							cutpoint.dAccuracyER = cutpoint.dAccuracyER *0.368/cutpoint.iNumPerformances;
		 
							cutpoint.dBrierScoreE0 = cutpoint.dBrierScoreE0 *0.632/cutpoint.iNumPerformances;
							cutpoint.dBrierScoreER = cutpoint.dBrierScoreER *0.368/cutpoint.iNumPerformances;
							
						 	if (bOutputToFile){
								
								cutpoint.dRootMeanSquareErrorE0 = cutpoint.dRootMeanSquareErrorE0 *0.632/cutpoint.iNumPerformances;
								cutpoint.dRootMeanSquareErrorER = cutpoint.dRootMeanSquareErrorER *0.368/cutpoint.iNumPerformances;
			
								cutpoint.dRelativeAbsoluteErrorE0 = cutpoint.dRelativeAbsoluteErrorE0 *0.632/cutpoint.iNumPerformances;
								cutpoint.dRelativeAbsoluteErrorER = cutpoint.dRelativeAbsoluteErrorER *0.368/cutpoint.iNumPerformances;
			
								cutpoint.dRootMeanPriorSquaredErrorE0 = cutpoint.dRootMeanPriorSquaredErrorE0 *0.632/cutpoint.iNumPerformances;
								cutpoint.dRootMeanPriorSquaredErrorER = cutpoint.dRootMeanPriorSquaredErrorER *0.368/cutpoint.iNumPerformances;
			
								cutpoint.dRootRelativeSquaredErrorE0 = cutpoint.dRootRelativeSquaredErrorE0 *0.632/cutpoint.iNumPerformances;
								cutpoint.dRootRelativeSquaredErrorER = cutpoint.dRootRelativeSquaredErrorER *0.368/cutpoint.iNumPerformances;
			
								cutpoint.dKBRelatieveInformationE0 = cutpoint.dKBRelatieveInformationE0 *0.632/cutpoint.iNumPerformances;
								cutpoint.dKBRelatieveInformationER = cutpoint.dKBRelatieveInformationER *0.368/cutpoint.iNumPerformances;
			
								cutpoint.dKBInformationE0 = cutpoint.dKBInformationE0 *0.632/cutpoint.iNumPerformances;
								cutpoint.dKBInformationER = cutpoint.dKBInformationER *0.368/cutpoint.iNumPerformances;
			
								cutpoint.dKBMeanInformationE0 = cutpoint.dKBMeanInformationE0 *0.632/cutpoint.iNumPerformances;
								cutpoint.dKBMeanInformationER = cutpoint.dKBMeanInformationER *0.368/cutpoint.iNumPerformances;
			
								cutpoint.dKappaE0 = cutpoint.dKappaE0 *0.632/cutpoint.iNumPerformances;
								cutpoint.dKappaER = cutpoint.dKappaER *0.368/cutpoint.iNumPerformances;
							 }
							for (int i =0;i<m_numClasses;i++){
								cutpoint.dTPE0[i] = cutpoint.dTPE0[i] *0.632/cutpoint.iNumPerformances;
								cutpoint.dTPER[i] = cutpoint.dTPER[i] *0.368/cutpoint.iNumPerformances;
								 if (bOutputToFile){
									if (cutpoint.iNumRocE0[i]> 0.0){
										cutpoint.dROCE0[i] = cutpoint.dROCE0[i] *0.632/cutpoint.iNumRocE0[i];
									} else {cutpoint.dROCE0[i]=0; }
									if (cutpoint.iNumRocER[i]> 0.0){
										cutpoint.dROCER[i] = cutpoint.dROCER[i] *0.368/cutpoint.iNumRocER[i];
									} else {cutpoint.dROCER[i]=0; }
															
									cutpoint.dFmeasureE0[i] = cutpoint.dFmeasureE0[i] *0.632/cutpoint.iNumPerformances;
									cutpoint.dFmeasureER[i] = cutpoint.dFmeasureER[i] *0.368/cutpoint.iNumPerformances;
									
									cutpoint.dRecallE0[i] = cutpoint.dRecallE0[i] *0.632/cutpoint.iNumPerformances;
									cutpoint.dRecallER[i] = cutpoint.dRecallER[i] *0.368/cutpoint.iNumPerformances;
									
									cutpoint.dPrecisionE0[i] = cutpoint.dPrecisionE0[i] *0.632/cutpoint.iNumPerformances;
									cutpoint.dPrecisionER[i] = cutpoint.dPrecisionER[i] *0.368/cutpoint.iNumPerformances;
								
									dMeanROC = dMeanROC + cutpoint.dROCE0[i]+ cutpoint.dROCER[i];
									dMeanFmeasure = dMeanFmeasure + cutpoint.dFmeasureE0[i]+ cutpoint.dFmeasureER[i];
									dMeanRecall = dMeanRecall + cutpoint.dRecallE0[i]+ cutpoint.dRecallER[i];
									dMeanPrecision = dMeanPrecision + cutpoint.dPrecisionE0[i]+ cutpoint.dPrecisionER[i];
								
								 }
								dMeanTP = dMeanTP + cutpoint.dTPE0[i]+ cutpoint.dTPER[i];
							}
							
							dMeanTP = dMeanTP/m_numClasses;dMeanROC = dMeanROC/m_numClasses;dMeanFmeasure = dMeanFmeasure/m_numClasses;dMeanRecall = dMeanRecall/m_numClasses;dMeanPrecision = dMeanPrecision/m_numClasses;
							
							Collections.sort(cutpoint.listFSS);
							 String sFSS1 ="-",sFSS2 ="-";
							 int iFSS1 =0,iFSS2=0;
							 double dFSS1merit =0, dFSS2merit =0;
							 ArrayList arrTempFSSlist =cutpoint.listFSS;
							 if (arrTempFSSlist.size()>0) {
								 sFSS1 =((SubsetConsensual)arrTempFSSlist.get(0)).getSubsetNames();
								 iFSS1 =((SubsetConsensual)arrTempFSSlist.get(0)).getSelectedCounts();
								 dFSS1merit =((SubsetConsensual)arrTempFSSlist.get(0)).getMerit();
							 }
		
							 if (arrTempFSSlist.size()>1) {
								 sFSS2 =((SubsetConsensual)arrTempFSSlist.get(1)).getSubsetNames();
								 iFSS2 =((SubsetConsensual)arrTempFSSlist.get(1)).getSelectedCounts();
								 dFSS2merit =((SubsetConsensual)arrTempFSSlist.get(1)).getMerit();
							 }
							
							 if (cutBestCutpoint==null) {
									cutBestCutpoint = cutpoint;
									if (flagWrapperAccuracy) {
										dMaxPerformance = cutpoint.dAccuracyE0+cutpoint.dAccuracyER;
									} else {
										dMaxPerformance = dMeanTP; // If bestcutpoint is the one with best MeanTP
									}
								} else {
									if (flagWrapperAccuracy) {
										if (dMaxPerformance<(cutpoint.dAccuracyE0+cutpoint.dAccuracyER)) {
											cutBestCutpoint = cutpoint;
											dMaxPerformance = cutpoint.dAccuracyE0+cutpoint.dAccuracyER; //If bestcutpoint is the one with best MeanTP
										}
									} else {
										if (dMaxPerformance<dMeanTP) {
											cutBestCutpoint = cutpoint;
											dMaxPerformance = dMeanTP; //If bestcutpoint is the one with best MeanTP
										}
									}
									
								}
							 if (bOutputToFile){
								 for (int i=0;i<m_numClasses;i++) cutpoint.arrClassPriors[i]=cutpoint.arrClassPriors[i]/cutpoint.iNumPerformances;
									
								cutpoint.dFayyadEntropy = cutpoint.dFayyadEntropy/cutpoint.iNumPerformances;
								 
								 sCutPoints="";sIns="";sTP="";sROC="";sFmeasure="";sRecall="";sPrecision="";
								 for (int i =0;i<m_numClasses;i++){
									 if (i<m_numClasses-1) sCutPoints=sCutPoints+cutpoint.listPoints.get(i)+",";
									 sIns=sIns+(Math.rint(cutpoint.arrClassPriors[i]*10)/10)+",";
									 sTP=sTP + ","+(Math.rint((cutpoint.dTPE0[i]+cutpoint.dTPER[i])*100)/100);
									 sROC=sROC + ","+(Math.rint((cutpoint.dROCE0[i]+cutpoint.dROCER[i])*100)/100);
									 sFmeasure=sFmeasure +","+(Math.rint((cutpoint.dFmeasureE0[i]+cutpoint.dFmeasureER[i])*100)/100);
									 sRecall=sRecall + ","+(Math.rint((cutpoint.dRecallE0[i]+cutpoint.dRecallER[i])*100)/100);
									 sPrecision=sPrecision + ","+(Math.rint((cutpoint.dPrecisionE0[i]+cutpoint.dPrecisionER[i])*100)/100);
								 }
								 
								 bw.write(sCutPoints+sIns+(Math.rint(cutpoint.dFayyadEntropy*1000)/1000)+sTP+","+(Math.rint(dMeanTP*100)/100)+","+cutpoint.iNumPerformances+","+(Math.rint((cutpoint.dAccuracyE0+cutpoint.dAccuracyER)*100)/100)+","+(Math.rint((cutpoint.dBrierScoreE0+cutpoint.dBrierScoreER)*1000)/1000)+sROC+","+(Math.rint(dMeanROC*100)/100)+sFmeasure+","+(Math.rint(dMeanFmeasure*100)/100)+sRecall+","+(Math.rint(dMeanRecall*100)/100)+sPrecision+","+(Math.rint(dMeanPrecision*100)/100)+","+(Math.rint((cutpoint.dRootMeanSquareErrorE0+cutpoint.dRootMeanSquareErrorER)*100)/100)+","+(Math.rint((cutpoint.dRelativeAbsoluteErrorE0+cutpoint.dRelativeAbsoluteErrorER)*100)/100)+","+(Math.rint((cutpoint.dRootMeanPriorSquaredErrorE0+cutpoint.dRootMeanPriorSquaredErrorER)*100)/100)+","+(Math.rint((cutpoint.dRootRelativeSquaredErrorE0+cutpoint.dRootRelativeSquaredErrorER)*100)/100)+","+(Math.rint((cutpoint.dKBRelatieveInformationE0+cutpoint.dKBRelatieveInformationER)*100)/100)+","+(Math.rint((cutpoint.dKBInformationE0+cutpoint.dKBInformationER)*100)/100)+","+(Math.rint((cutpoint.dKBMeanInformationE0+cutpoint.dKBMeanInformationER)*100)/100)+","+(Math.rint((cutpoint.dKappaE0+cutpoint.dKappaER)*100)/100)+","+sFSS1+","+iFSS1+","+(Math.rint(dFSS1merit*1000)/1000)+","+sFSS2+","+iFSS2+","+(Math.rint(dFSS2merit*1000)/1000)+"\n");
							 }
							 dMeanTP=0;dMeanROC=0;dMeanFmeasure=0;dMeanRecall=0;dMeanPrecision=0;
								
					 }	
				 }
				 if (bOutputToFile){
					 bw.close();
				 }
					 // Returns the train and test set discretized with the best Mean(TP)
			 } catch (Exception ex){
				 ex.printStackTrace();
			 }  
			 
			 Instances insTestP=null;
			 if (acTrainTestPair.insTest!=null) insTestP = new Instances(acTrainTestPair.insTest);
			 rtnTrainTestPair = DiscretizeFeature(new TrainTestPair(new Instances(acTrainTestPair.insTrain), insTestP),cutBestCutpoint.listPoints,acTrainTestPair.insTrain.numAttributes()-1);
			 
		 } else {  // The class is nominal
			 Instances insTestP=null;
			 if (acTrainTestPair.insTest!=null) insTestP = new Instances(acTrainTestPair.insTest);
			 rtnTrainTestPair = new TrainTestPair(new Instances(acTrainTestPair.insTrain),insTestP);
		 }
		 // Discretization of the original train test set with the "better" cutpoint
		 
		 rtnTrainTestPair = supervisedDiscretization(rtnTrainTestPair);
		 
		 if (flagPerformPrediction) {
			 ModelValidation(rtnTrainTestPair.insTrain,dataSetToPredict);		 
		 }
		 
		 return rtnTrainTestPair;
	 }		
	 
	 // Avoids to compute extremely imbalanced combinations that are useless
	 // Calculates intervals depending on number of cutpoints for class
	 private static void ValidCutPointCombinations() {
		 RecursiveValidCutPointCombinations(m_numClasses-1,iMinInstancesPerClass-1,new ArrayList());
	}
	 
	 private static void RecursiveValidCutPointCombinations(int iNumCuts,int actualCutIndex,ArrayList listCutPoints) {
		 if (iNumCuts > 0) {	
				 ArrayList listCutPointsTemp;
				 int iMaxI =m_CutPoints.length-(iMinInstancesPerClass*(iNumCuts));
				 
					
				 while (actualCutIndex<=iMaxI) { // No estoy seguro de si sobra el =
					 listCutPointsTemp = new ArrayList(listCutPoints);
					 listCutPointsTemp.add(new Double(m_CutPoints[actualCutIndex]));
					 
				     if (iNumCuts == 1) {
					 	m_CutPointCombinations.add(new CutPointSetPredictor(listCutPointsTemp));
					 	//System.out.println("MaxI:"+iMaxI+"; CutPoints: "+listCutPointsTemp.toString());
				     }
				     
				     RecursiveValidCutPointCombinations(iNumCuts-1,actualCutIndex+iMinInstancesPerClass,listCutPointsTemp); // Sends a deep copy of the arraylist
				     
				     actualCutIndex++;
				 }
			  }
     }
	 

	private static EvaluationBrier ModelValidation(Instances train, Instances test){
		 EvaluationBrier evalC=null;
		 try {
//			 Fayyad and Consensual CFS
			 	ArrayList arrCfs = new ArrayList();
				// If the cutpoint exist in the list the lists of previous CFS are passed to be actualized
				//TrainTestPair classDisTrain = ConsensualCFS(new TrainTestPair(train,test),arrCfs,false);						
			 	TrainTestPair classDisTrain = new TrainTestPair(train,test);						
				
			 	Classifier m_Classifier = getNaiveBayes();
				
				//classDisTrain.insTrain.setClassIndex(classDisTrain.insTrain.numAttributes()-1);
				classDisTrain.insTrain.setClassIndex(classDisTrain.insTrain.numAttributes()-1);
				try {
					m_Classifier.buildClassifier(classDisTrain.insTrain);
				} catch (Exception e) {
					System.out.println("Error in model validation");	
					e.printStackTrace();
				}    
				
				// Evaluation
				evalC = new EvaluationBrier(classDisTrain.insTrain);
				if (flagPerformPrediction) {}
				try {
					evalC.evaluateModel(m_Classifier, classDisTrain.insTest);
				} catch (Exception e) {
					System.out.println("Error evaluating in model validation: ");	
					e.printStackTrace();
				}
				
				if (flagPerformPrediction) {
					
					 Date actualDate = new Date();
					 String sFichero = "Prediction_LearningFile_"+sLearningFile+"_FileToPredict_"+sTopredictFile+"_NaiveBayes.csv";
					 File fichero = new File(sFichero);
					 if (fichero.exists()) fichero.delete();
					 BufferedWriter bw = new BufferedWriter(new FileWriter(sFichero));
					 
					 FastVector actualPredictions= evalC.predictions();
					 String sClassValues ="";
					 Enumeration enumClassValues;
					 Attribute classAtt = dataSetToPredict.classAttribute();
					 
					 for (int i=0; i<classAtt.numValues(); i++) {
						 sClassValues = sClassValues+","+classAtt.value(i);
					 }
					 
					 bw.write("Actual,Predicted"+sClassValues+"\n");
					 NominalPrediction actualPrediction =null;
					 String sPredictionDistribution ="";
					 String realRec ="?";
					 double[] arrPredicDist;
					  for (int i=0; i<actualPredictions.size();i++){
						  actualPrediction =(NominalPrediction) actualPredictions.elementAt(i);
						 arrPredicDist=actualPrediction.distribution();
						 for (int j=0; j<arrPredicDist.length;j++) {
							 sPredictionDistribution=sPredictionDistribution+","+Math.rint(arrPredicDist[j]*10000)/10000;
						 } 
						 if (!Double.isNaN(actualPrediction.actual())) {
							 realRec = classAtt.value((int)actualPrediction.actual()); 							 
						 }
						 
						 bw.write(realRec+","+classAtt.value((int) actualPrediction.predicted())+sPredictionDistribution+"\n");
						 sPredictionDistribution ="";
					 }
					 
					 bw.close();
					
				}
				

		 } catch (Exception ex){
			 ex.printStackTrace();
		 }   
		 return evalC; 
	 }	
	 
	 /*It only uses train, test is for do to it the same transformations*/
	 private static TrainTestPair ConsensualCFS(TrainTestPair datasetPair, ArrayList listSubsets, boolean bOutputToFile){
		 // Using: Leave one out (LOO)
		
		 int numInstances = datasetPair.insTrain.numInstances();
		 datasetPair.insTrain.setClassIndex(datasetPair.insTrain.numAttributes()-1);
		 Instances fssInstances = new Instances(datasetPair.insTrain);
		 // Cfs initializations
		 AttributeSelectionMerit attsel2= new AttributeSelectionMerit();
		 CfsSubsetEval eval2 = new CfsSubsetEval();
		 BestFirstMulti search2 = new BestFirstMulti();
		 attsel2.setEvaluator(eval2);
		 attsel2.setSearch(search2);
		

		 String sSelectedAtt = "";
		 String sSelectedIndexes ="";
		 double dMerit2 =0;
		 SubsetConsensual actualSubsetConsensual;
		 weka.filters.unsupervised.attribute.Remove removeData = new weka.filters.unsupervised.attribute.Remove();
		 int [] indicesFull2 = null;
		 Instances fssInstancesTest = null;
		 
		
		 // Leaving one out (LOO)
		 for (int iActualIns =0; iActualIns < numInstances; iActualIns++) {
			fssInstances = new Instances(datasetPair.insTrain);
			fssInstances.delete(iActualIns);
			 
			fssInstances.setClassIndex(fssInstances.numAttributes()-1);
			
			TrainTestPair fssInstancesPair =supervisedDiscretization(new TrainTestPair(new Instances(fssInstances), null));
			fssInstances = fssInstancesPair.insTrain;
			
			fssInstances.setClassIndex(fssInstances.numAttributes()-1);
			if (fssInstances.numAttributes()>1) {
				// Fss
				try {
					attsel2.SelectAttributes(fssInstances);
					sSelectedIndexes =attsel2.selectedAttributes().toString();
				 
					dMerit2 = attsel2.m_bestMerit;
					
					indicesFull2 = attsel2.selectedAttributes();
					sSelectedAtt = "";
					for (int iAtt=0; iAtt<indicesFull2.length-1;iAtt++) {
						sSelectedAtt = sSelectedAtt+ fssInstances.attribute(indicesFull2[iAtt]).name()+ ";";
					}
					
					// listPredictorsDiscretization is actualized in supervisedDiscretization fuction
					actualSubsetConsensual =new SubsetConsensual(sSelectedAtt,indicesFull2,1,dMerit2,listPredictorsDiscretization);
					// System.out.println("Num features before CFS:"+fssInstances.numAttributes()+", after: "+indicesFull2.length);
					 
					 if (listSubsets.contains(actualSubsetConsensual)) {
						 SubsetConsensual actualizeSubsetConsensual = (SubsetConsensual)listSubsets.get(listSubsets.indexOf(actualSubsetConsensual));
						 actualizeSubsetConsensual.setSelectedCounts(actualizeSubsetConsensual.getSelectedCounts()+1);
						 actualizeSubsetConsensual.setMerit(actualizeSubsetConsensual.getMerit()+actualSubsetConsensual.getMerit());
						 
						 // Check if the discretization exist to add it or actualize counts
						  for (int i=0;i<indicesFull2.length-1;i++){
							 if (actualizeSubsetConsensual.arrDiscretizationPoints[i].contains(actualSubsetConsensual.arrDiscretizationPoints[i].get(0))){
								 ((CutPointSetPredictor)(actualizeSubsetConsensual.arrDiscretizationPoints[i].get(actualizeSubsetConsensual.arrDiscretizationPoints[i].indexOf(actualSubsetConsensual.arrDiscretizationPoints[i].get(0))))).countSelected++;
							 } else {
								 actualizeSubsetConsensual.arrDiscretizationPoints[i].add(actualSubsetConsensual.arrDiscretizationPoints[i].get(0));
							 }
						 }
						 
						 
					 } else {
						 listSubsets.add(actualSubsetConsensual);
					 }
				 } catch (Exception ex){
					 ex.printStackTrace();
					 
				 } 
			}
		 }
		 
		 Collections.sort(listSubsets);
		 
		 SubsetConsensual bestSubsetConsensual=null;
		 // Output results to file
		 try {
			 if (bOutputToFile){
				 String sClassValues ="";
				 Enumeration enumClassValues = datasetPair.insTrain.classAttribute().enumerateValues();
				 while( enumClassValues.hasMoreElements() )
					 sClassValues = sClassValues +"____" + (( String )enumClassValues.nextElement());
				 
				String sFichero = "CFSsubsetConsensualScores_"+sLearningFile+sClassValues+".csv";
				File fichero = new File(sFichero);
				if (fichero.exists()) fichero.delete();
				BufferedWriter bw = new BufferedWriter(new FileWriter(sFichero));
				bw.write("SubsetNames, SelectedCounts, Merit, MeanMerit\n");
				
				Iterator iter = listSubsets.iterator();
				while (iter.hasNext()) {
					SubsetConsensual recordSubsetConsensual = (SubsetConsensual) iter.next();
					bw.write(recordSubsetConsensual.getSubsetNames()+"," + recordSubsetConsensual.getSelectedCounts()+ ","+recordSubsetConsensual.getMerit()+","+recordSubsetConsensual.getMerit()/recordSubsetConsensual.getSelectedCounts()+"\n");
				}
				bw.close();
			 }
			 boolean bSetFound = false;
			 try {
				 bestSubsetConsensual = (SubsetConsensual)listSubsets.get(0);
				 bSetFound = true;
			 } catch(Exception e){
				// e.printStackTrace();
				 bSetFound = false;
			 }
			if (bSetFound)	removeData.setAttributeIndicesArray(FindSubsetIndex(bestSubsetConsensual.getSubsetNames(),datasetPair.insTrain));
			else {
				int[] arrFeature= {datasetPair.insTrain.numAttributes()-1};
				removeData.setAttributeIndicesArray(arrFeature);
			}
		 	removeData.setInvertSelection(true);
			removeData.setInputFormat(datasetPair.insTrain);
			fssInstances =Filter.useFilter(datasetPair.insTrain, removeData);
			if (datasetPair.insTest!=null) {
				//removeData.setInputFormat(datasetPair.insTest);
				fssInstancesTest =Filter.useFilter(datasetPair.insTest, removeData);
			}
		 
		 } catch (Exception e) {
			 //e.printStackTrace();
		}
		 
		 // Cambiar por discretizacin ms votada
		 return PerformPredictorsBestDiscretization(new TrainTestPair(fssInstances, fssInstancesTest),bestSubsetConsensual);
	 }
	
	 private static TrainTestPair PerformPredictorsBestDiscretization(TrainTestPair trainTestPair,SubsetConsensual bestSubsetConsensual) {
		 //ArrayList[] arrFeatureCutPoints =bestSubsetConsensual.arrDiscretizationPoints;
		 CutPointSetPredictor bestCutPoints;
		 if(bestSubsetConsensual!=null)
			 for (int i=0;i<bestSubsetConsensual.arrDiscretizationPoints.length;i++) {
				 Collections.sort(bestSubsetConsensual.arrDiscretizationPoints[i]);
				 bestCutPoints = (CutPointSetPredictor)bestSubsetConsensual.arrDiscretizationPoints[i].get(0);
				 trainTestPair = DiscretizeFeature(trainTestPair,bestCutPoints.listPoints,i);
			 }
		  
		return trainTestPair;
	}

	private static TrainTestPair DiscretizeFeature(TrainTestPair trainTestPair, ArrayList listPoints, int iAttributeIndex) {
		double actualValue,cutPointToPerform,previousCutPoint;
		boolean notFound;
		int iActualCutPoint;
		String sNominalValues;
		
		 weka.filters.unsupervised.attribute.Add attClassDis = new weka.filters.unsupervised.attribute.Add(); 
		 
		//Introduce the attribute header
			// There must be at least one cutpoint
			 previousCutPoint =((Double)listPoints.get(0)).doubleValue();
			 sNominalValues = "Less_"+previousCutPoint+",";
			 
			 iActualCutPoint =1;
			 if (listPoints.size()>1){
				 while (iActualCutPoint<listPoints.size()) {
					 cutPointToPerform =((Double)listPoints.get(iActualCutPoint)).doubleValue();
					 sNominalValues = sNominalValues+ previousCutPoint +"__"+ cutPointToPerform+",";
					 previousCutPoint = cutPointToPerform;
					 iActualCutPoint ++;
				}
			 }
		cutPointToPerform =((Double)listPoints.get(listPoints.size()-1)).doubleValue();
		sNominalValues = sNominalValues+ "More_"+cutPointToPerform;
		
		 attClassDis.setAttributeName(trainTestPair.insTrain.attribute(iAttributeIndex).name());
		 attClassDis.setNominalLabels(sNominalValues);
		 attClassDis.setAttributeIndex((new Integer(iAttributeIndex+2)).toString());
		 try {
			attClassDis.setInputFormat(trainTestPair.insTrain);
			trainTestPair.insTrain=  Filter.useFilter(trainTestPair.insTrain, attClassDis);
			if (trainTestPair.insTest != null)
				trainTestPair.insTest=  Filter.useFilter(trainTestPair.insTest, attClassDis);
		 } catch (Exception e) {
				e.printStackTrace();
			}
		
		 
		// Discretize Train
		 for (int iActualIns=0;iActualIns<trainTestPair.insTrain.numInstances();iActualIns++) {
			 actualValue =trainTestPair.insTrain.instance(iActualIns).value(iAttributeIndex);
			 
			 // There must be at least one cutpoint
			 previousCutPoint =((Double)listPoints.get(0)).doubleValue();
			 if (actualValue<=previousCutPoint) {
				 trainTestPair.insTrain.instance(iActualIns).setValue(iAttributeIndex+1, "Less_"+previousCutPoint);
			 } else {
				 notFound = true;
				 iActualCutPoint =1;
	
				 if (listPoints.size()>1){
					 while (notFound && (iActualCutPoint<listPoints.size())) {
						 cutPointToPerform =((Double)listPoints.get(iActualCutPoint)).doubleValue();
						 if (actualValue<=cutPointToPerform) {
							 trainTestPair.insTrain.instance(iActualIns).setValue(iAttributeIndex+1, previousCutPoint +"__"+ cutPointToPerform);
							 notFound =false;
						 } 
						 previousCutPoint = cutPointToPerform;
						 iActualCutPoint ++;
					 }
				 } 
				 if (notFound) {
					 cutPointToPerform =((Double)listPoints.get(listPoints.size()-1)).doubleValue();
					 if (actualValue>cutPointToPerform) {
						 trainTestPair.insTrain.instance(iActualIns).setValue(iAttributeIndex+1, "More_"+cutPointToPerform);
					 } 
				 }
			 }
		 }
		
		// Discretize test
		 if (trainTestPair.insTest != null) {
			 for (int iActualIns=0;iActualIns<trainTestPair.insTest.numInstances();iActualIns++) {
				 actualValue =trainTestPair.insTest.instance(iActualIns).value(iAttributeIndex);
				 
				 // There must be at least one cutpoint
				 previousCutPoint =((Double)listPoints.get(0)).doubleValue();
				 if (actualValue<=previousCutPoint) {
					 trainTestPair.insTest.instance(iActualIns).setValue(iAttributeIndex+1, "Less_"+previousCutPoint);
				 } else {
					 notFound = true;
					 iActualCutPoint =1;
		
					 if (listPoints.size()>1){
						 while (notFound && (iActualCutPoint<listPoints.size())) {
							 cutPointToPerform =((Double)listPoints.get(iActualCutPoint)).doubleValue();
							 if (actualValue<=cutPointToPerform) {
								 trainTestPair.insTest.instance(iActualIns).setValue(iAttributeIndex+1, previousCutPoint +"__"+ cutPointToPerform);
								 notFound =false;
							 } 
							 previousCutPoint = cutPointToPerform;
							 iActualCutPoint ++;
						 }
					 } 
					 if (notFound) {
						 cutPointToPerform =((Double)listPoints.get(listPoints.size()-1)).doubleValue();
						 if (actualValue>cutPointToPerform) {
							 trainTestPair.insTest.instance(iActualIns).setValue(iAttributeIndex+1, "More_"+cutPointToPerform);
						 } 
					 }
				 }
			 }
		
		 }
				
		 // Remove the prediscretization column
		 	weka.filters.unsupervised.attribute.Remove removeData = new weka.filters.unsupervised.attribute.Remove();
			int[] arrClass = {iAttributeIndex};
			removeData.setAttributeIndicesArray(arrClass);
			try {
				removeData.setInputFormat(trainTestPair.insTrain);
			
				trainTestPair.insTrain = Filter.useFilter(trainTestPair.insTrain, removeData);
				 if (trainTestPair.insTest != null) 
						trainTestPair.insTest = Filter.useFilter(trainTestPair.insTest, removeData);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		return trainTestPair;
	}

	private static int[] FindSubsetIndex(String subsetNames, Instances insTrain) {
		StringTokenizer st = new StringTokenizer(subsetNames,";");
	    int iNumFeatures =st.countTokens();
		int [] arrFeatures = new int[iNumFeatures+1];
		for (int i=0;i<iNumFeatures;i++)
	    	 arrFeatures[i]= insTrain.attribute(st.nextToken()).index();
		arrFeatures[iNumFeatures]=insTrain.numAttributes()-1;
	    return arrFeatures;
	}

	/*It only uses train, test is for doing to it the same transformations*/
	 private static TrainTestPair supervisedDiscretization(TrainTestPair dataset){
		 TrainTestPair discretizedInstances = null;
		 try {

			 dataset.insTrain.setClassIndex(dataset.insTrain.numAttributes()-1);
			 if (dataset.insTest!=null) dataset.insTest.setClassIndex(dataset.insTest.numAttributes()-1);
			 if (dataSetToPredict!=null) dataSetToPredict.setClassIndex(dataSetToPredict.numAttributes()-1);

			 DiscretizeFayyad discretizeData = new DiscretizeFayyad();
			 discretizeData.setInputFormat(dataset.insTrain);
			 
			 dataset.insTrain = Filter.useFilter(dataset.insTrain, discretizeData);
			 m_Fayyad_entropy = discretizeData.m_entropy;
			 
			 //Delete attributes set in an unique interval
			 ArrayList arrToKeep = new ArrayList();
			 for (int i=0; i<dataset.insTrain.numAttributes();i++) {
				 if (dataset.insTrain.attribute(i).numValues()>1) arrToKeep.add(new Integer(i));					
			 }
			 int[] arrInToKeep= new int[arrToKeep.size()];
			 listPredictorsDiscretization = new ArrayList[arrToKeep.size()-1];
			 for (int i=0; i<arrToKeep.size()-1; i++){ 
				 arrInToKeep[i]= ((Integer)arrToKeep.get(i)).intValue();
				 listPredictorsDiscretization[i] = new ArrayList();
				 listPredictorsDiscretization[i].add(new CutPointSetPredictor(discretizeData.m_CutPoints[arrInToKeep[i]]));
			 }				 
			 arrInToKeep[arrToKeep.size()-1]= ((Integer)arrToKeep.get(arrToKeep.size()-1)).intValue();
			 
			 weka.filters.unsupervised.attribute.Remove removeData = new weka.filters.unsupervised.attribute.Remove();
			 removeData.setInvertSelection(true);
			 removeData.setAttributeIndicesArray(arrInToKeep);
			 removeData.setInputFormat(dataset.insTrain);
			 dataset.insTrain = Filter.useFilter(dataset.insTrain, removeData);
			 
			 if (dataset.insTest!=null){
				dataset.insTest = Filter.useFilter(dataset.insTest, discretizeData);
				dataset.insTest = Filter.useFilter(dataset.insTest, removeData);
			 }
			 
			 if (dataSetToPredict!=null){
				 dataSetToPredict = Filter.useFilter(dataSetToPredict, discretizeData);
				 dataSetToPredict = Filter.useFilter(dataSetToPredict, removeData);
			 }
			 
			 discretizedInstances = new TrainTestPair(dataset.insTrain,dataset.insTest);
	
		 } catch (Exception e) {
				e.printStackTrace();
		 }
		 
		 return discretizedInstances;
	 }
	 
	 /* For each call it actualizes the list of cutpoints and their associated list of performances*/
	 private static void ClassDiscretization(TrainTestPair datasetPair, int iIteration, boolean bFileOutput){
		 TrainTestPair discretizedClassInstances = null;
		 Instances datasetTrain= new Instances(datasetPair.insTrain); 
		 
		 try {
			 m_NumAtributes = datasetTrain.numAttributes();
			 int m_ClassIndex =m_NumAtributes-1;
			
			 datasetTrain.sort(datasetTrain.attribute(m_ClassIndex));
			 
			 //m_CutPoints = cutPointsForSubset(datasetTrain, m_ClassIndex, 0, datasetTrain.numInstances(),iIteration);
			  
			 // Begin to evaluate all valid cutpoints (extremely umbalanced ones can be avoided)
			 Iterator iter = m_CutPointCombinations.iterator();
			 CutPointSetPredictor actualCutPointSet;
			 CutPointSet cutpoint;
			 while (iter.hasNext()) {
				 actualCutPointSet = (CutPointSetPredictor)iter.next();
				 
				 TrainTestPair classDisTrain = DiscretizeFeature(new TrainTestPair(new Instances(datasetPair.insTrain), new Instances(datasetPair.insTest)),actualCutPointSet.listPoints,m_ClassIndex);
					
				 cutpoint = new CutPointSet(m_numClasses); // Estoy duplicando infomacin?
				 cutpoint.listPoints = new ArrayList(actualCutPointSet.listPoints);
			
				 boolean bListHasCutPoint = false;
						if (listCutPoints.contains(cutpoint)) bListHasCutPoint= true;
						
						// Begin posible nueva funcion
						// Fayyad and Consensual CFS
						// If the cutpoint exist in the list the lists of previous CFS are passed to be actualized
						if (bListHasCutPoint) {
							classDisTrain = ConsensualCFS(new TrainTestPair(new Instances(classDisTrain.insTrain), new Instances(classDisTrain.insTest)),((CutPointSet)listCutPoints.get(listCutPoints.indexOf(cutpoint))).listFSS,false);
						} else {
							classDisTrain = ConsensualCFS(new TrainTestPair(new Instances(classDisTrain.insTrain), new Instances(classDisTrain.insTest)),cutpoint.listFSS,false);						
						}
						
						if (classDisTrain!=null){ 
						
							Classifier m_Classifier = getNaiveBayes();
							//classDisTrain.insTrain.setClassIndex(classDisTrain.insTrain.numAttributes()-1);
							classDisTrain.insTrain.setClassIndex(classDisTrain.insTrain.numAttributes()-1);
							try {
								m_Classifier.buildClassifier(classDisTrain.insTrain);
							} catch (Exception e) {
								System.out.println("Error in Boot: "+iIteration+ ".");	
								e.printStackTrace();
							}    
							// Evaluation
							EvaluationBrier evalC = new EvaluationBrier(classDisTrain.insTrain);
							try {
								evalC.evaluateModel(m_Classifier, classDisTrain.insTest);
							} catch (Exception e) {
								System.out.println("Error evaluating in Boot: "+iIteration+ ".");	
								e.printStackTrace();
							}   
							EvaluationBrier evalCR = new EvaluationBrier(classDisTrain.insTrain);
							evalCR.evaluateModel(m_Classifier, classDisTrain.insTrain);
							
							// End posible funcin extena
							
							/*Add cutpoint if it do not exist and add performance to the list*/
							if (bListHasCutPoint) { // Actulize performance of the cutpoint in the list
								ActualizeCutPointPerformance((CutPointSet)listCutPoints.get(listCutPoints.indexOf(cutpoint)), evalC,evalCR,bFileOutput);
							 } else {  // Adds the cut point to the list
								 listCutPoints.add(ActualizeCutPointPerformance(cutpoint, evalC,evalCR,bFileOutput));
							 }
						}
					// }  // End if test balanced
				 
			} 			 // End_while to evaluate all possible 3 cutpoints
		} catch (Exception e) {e.printStackTrace();}

		
	 }
	 
	 private static void ClassDiscretizationFilter(TrainTestPair datasetPair, int iIteration, boolean bFileOutput){
		 TrainTestPair discretizedClassInstances = null;
		 Instances datasetTrain= new Instances(datasetPair.insTrain); 
		 
		 try {
			 m_NumAtributes = datasetTrain.numAttributes();
			 int m_ClassIndex =m_NumAtributes-1;
			
			  
			 // Begin to evaluate all valid cutpoints (extremely umbalanced ones can be avoided)
			 Iterator iter = m_CutPointCombinations.iterator();
			 CutPointSetPredictor actualCutPointSet;
			 CutPointSet cutpoint;
			 Instances insTestP=null;
			 if (datasetPair.insTest!=null) insTestP = new Instances(datasetPair.insTest);
			 
			 while (iter.hasNext()) {
				 actualCutPointSet = (CutPointSetPredictor)iter.next();
				 
				 TrainTestPair classDisTrain = DiscretizeFeature(new TrainTestPair(new Instances(datasetPair.insTrain), insTestP),actualCutPointSet.listPoints,m_ClassIndex);
					
				 cutpoint = new CutPointSet(m_numClasses); // Estoy duplicando infomacin?
				 cutpoint.listPoints = new ArrayList(actualCutPointSet.listPoints);
				 if (insTestP==null){
					 classDisTrain = ConsensualCFS(new TrainTestPair(new Instances(classDisTrain.insTrain), null),cutpoint.listFSS,false);
				 }
				 else {
					 classDisTrain = ConsensualCFS(new TrainTestPair(new Instances(classDisTrain.insTrain), new Instances(classDisTrain.insTest)),cutpoint.listFSS,false);
				 }
				 cutpoint.dFayyadEntropy = m_Fayyad_entropy;
				 listCutPoints.add(cutpoint);
						
			} 			 // End_while to evaluate all possible 3 cutpoints
		} catch (Exception e) {e.printStackTrace();}

		
	 }
	 
	 
	 static CutPointSet ActualizeCutPointPerformance(CutPointSet actualizeCutPoint, EvaluationBrier evalC, EvaluationBrier evalCR, boolean bFileOutput ) {
		
			actualizeCutPoint.dAccuracyE0+= evalC.pctCorrect();
			actualizeCutPoint.dAccuracyER+= evalCR.pctCorrect();
			
			try {
				actualizeCutPoint.dBrierScoreE0+= evalC.BrierScore();
				actualizeCutPoint.dBrierScoreER+= evalCR.BrierScore();
			} catch (Exception e) {}
			
		 if (bFileOutput) {
			 double []classPriors = evalC.getClassPriors();
			if (actualizeCutPoint.iNumPerformances==0) {
				for (int i=0;i<m_numClasses;i++) actualizeCutPoint.arrClassPriors=classPriors;
			} else {
				for (int i=0;i<m_numClasses;i++) actualizeCutPoint.arrClassPriors[i]+=classPriors[i];
			}
			
		
			for (int i =0;i<m_numClasses;i++){
					if (evalC.areaUnderROC(i)> 0.0){
					actualizeCutPoint.dROCE0[i]+= evalC.areaUnderROC(i);
					actualizeCutPoint.iNumRocE0[i]++;
				}
				if (evalCR.areaUnderROC(i)> 0.0){
					actualizeCutPoint.dROCER[i]+= evalCR.areaUnderROC(i);
					actualizeCutPoint.iNumRocER[i]++;
				}
				
				actualizeCutPoint.dFmeasureE0[i]+= evalC.fMeasure(i);
				actualizeCutPoint.dFmeasureER[i]+= evalCR.fMeasure(i);
				
				actualizeCutPoint.dRecallE0[i]+= evalC.recall(i);
				actualizeCutPoint.dRecallER[i]+= evalCR.recall(i);
				
				actualizeCutPoint.dPrecisionE0[i]+= evalC.precision(i);
				actualizeCutPoint.dPrecisionER[i]+= evalCR.precision(i);
			}
			actualizeCutPoint.dRootMeanSquareErrorE0+= evalC.rootMeanSquaredError();
			actualizeCutPoint.dRootMeanSquareErrorER+= evalCR.rootMeanSquaredError();
		

		
			try {
				actualizeCutPoint.dRelativeAbsoluteErrorE0+= evalC.relativeAbsoluteError();
				actualizeCutPoint.dRelativeAbsoluteErrorER+= evalCR.relativeAbsoluteError();
			} catch (Exception e) {}
					
			try {
				actualizeCutPoint.dRootMeanPriorSquaredErrorE0+= evalC.rootMeanPriorSquaredError();
				actualizeCutPoint.dRootMeanPriorSquaredErrorER+= evalCR.rootMeanPriorSquaredError();
			} catch (Exception e) {}
		
			try {
				actualizeCutPoint.dRootRelativeSquaredErrorE0+= evalC.rootRelativeSquaredError();
				actualizeCutPoint.dRootRelativeSquaredErrorER+= evalCR.rootRelativeSquaredError();
			} catch (Exception e) {}
			
			try {
				actualizeCutPoint.dKBRelatieveInformationE0+= evalC.KBRelativeInformation();
				actualizeCutPoint.dKBRelatieveInformationER+= evalCR.KBRelativeInformation();
			} catch (Exception e) {}
			
			try {
				actualizeCutPoint.dKBInformationE0+= evalC.KBInformation();
				actualizeCutPoint.dKBInformationER+= evalCR.KBInformation();
			} catch (Exception e) {}
			
			try {
				actualizeCutPoint.dKBMeanInformationE0+= evalC.KBMeanInformation();
				actualizeCutPoint.dKBMeanInformationER+= evalCR.KBMeanInformation();
			} catch (Exception e) {}
			
			try {
				actualizeCutPoint.dKappaE0+= evalC.kappa();
				actualizeCutPoint.dKappaER+= evalCR.kappa();
			} catch (Exception e) {}
			
		    actualizeCutPoint.dFayyadEntropy += m_Fayyad_entropy;
		}
		for (int i =0;i<m_numClasses;i++){
			actualizeCutPoint.dTPE0[i]+= evalC.truePositiveRate(i);
			actualizeCutPoint.dTPER[i]+= evalCR.truePositiveRate(i);
		}
		
		actualizeCutPoint.iNumPerformances++;
		return actualizeCutPoint;
	 }
	 
	 static String GetAttributeNames(int[] arrAttributes, Instances dataset){
		 String sResult = "";
		 
		 for (int i=0;i<arrAttributes.length-1;i++) 
			 sResult += dataset.attribute(arrAttributes[i]).name()+ ";";
		 
		 return sResult;
	 }
	 
	
	 static void ParseArguments(String[] args){
		 int i = 0, j;
	        String arg;
	        char flag;
	        
	        String outputfile = "";
	        outputfile = "";
	        
	        int numArgs = args.length;
	        
	        while (i < numArgs) {
		        if (args[i].startsWith("-")){	
		            arg = args[i];
		            i++;
	
		    // use this type of check for "wordy" arguments
		            if (arg.equals("-verbose")) {
		                System.out.println("verbose mode on");
		                vflag = true;
		            }
		            if (arg.equals("-Filter")) {
		                System.out.println("Filter class discretization on");
		                flagFilterClassDiscretization = true;
		                sFilter ="_filter_";
		            }
		            if (arg.equals("-WrapperAccuracy")) {
		                System.out.println("Wrapper accuracy driven class discretization on");
		                flagWrapperAccuracy = true;
		                sFilter ="_wrapperAccuracy_";
		            }
		            if (arg.equals("-WrapperMeanTP")) {
		                System.out.println("Wrapper mean(TP) driven class discretization on");
		                flagWrapperAccuracy = false;
		                sFilter ="_wrapperMeanTP_";
		            }
		            if (arg.equals("-ConsensualCFS")) {
		                System.out.println("ConsensualCFS on");
		                flagConsensualCFS = true;
		            }
		            if (arg.equals("-ModelBuilding")) {
		                System.out.println("ModelBuilding on");
		                flagModelBuilding = true;
		            }
		            if (arg.equals("-ModelValidation")) {
		                System.out.println("ModelValidation on");
		                flagModelValidation = true;
		            }
		            
		    // use this type of check for arguments that require arguments
		            else if (arg.equals("-boot")) {
		                if (i < args.length)
		                	m_NumBootstrap = Integer.parseInt(args[i++]);
		                else
		                    System.err.println("-boot requires a number");
		                if (vflag)
		                    System.out.println("number of bootstraps = " + m_NumBootstrap);
		            }
		            else if (arg.equals("-fold")) {
		                if (i < args.length)
		                	m_NumFolds = Integer.parseInt(args[i++]);
		                else
		                    System.err.println("-folds requires a number");
		                if (vflag)
		                    System.out.println("number of folds in cross-validation = " + m_NumFolds);
		            }
		            else if (arg.equals("-repeat")) {
		                if (i < args.length)
		                	m_NumCV = Integer.parseInt(args[i++]);
		                else
		                    System.err.println("-repeat requires a number");
		                if (vflag)
		                    System.out.println("number of times cross-validation is repeated = " + m_NumCV);
		            }
		            else if (arg.equals("-NumClasses")) {
		                if (i < args.length)
		                	m_numClasses = Integer.parseInt(args[i++]);
		                else
		                    System.err.println("-NumClasses requires a number");
		                if (vflag)
		                    System.out.println("number of intervals to discretize class = " + m_numClasses);
		            }
		            else if (arg.equals("-MinInstPerClass")) {
		                if (i < args.length)
		                	iMinInstancesPerClass = Integer.parseInt(args[i++]);
		                else
		                    System.err.println("-MinInstPerClass requires a number");
		                if (vflag)
		                    System.out.println("minimun number of instances per class (avoid extemely umbalanced class configurations and speed up model building and validation) = " + iMinInstancesPerClass);
		            }
		            else if (arg.equals("-LearningFile")) {
		                if (i < args.length) {
		                	sLearningFile = args[i++];
		                }
		                else {
		                    System.err.println("-LearningFile requires a file name in current directory");
		                }
		                if (vflag)
		                    System.out.println("Learning data filename = " + sLearningFile);
		            }
		            else if (arg.equals("-PerformPrediction")) {
		                if (i < args.length) {
		                	sTopredictFile = args[i++];
		                	flagPerformPrediction = true;
		                	numArgs--;
		                } else {
		                	System.err.println("-PerformPrediction requires a file name in current directory");
		                }
		                
		                if (vflag)
		                    System.out.println("Data to predict filename = " + sTopredictFile);
		            }
		            else if (arg.equals("-SetClassDiscretization")) {
		                if (i < args.length) {
		                	sExternalCutPoints = args[i++];
		                	flagSetClassDiscretization = true;
		                } else {
		                	System.err.println("-PerformPrediction requires a file name in current directory");
		                }
		                
		                if (vflag)
		                    System.out.println("Data to predict filename = " + sTopredictFile);
		            }
		            
		        }
	        }
	        if (i == args.length)
	            System.err.println("Usage: MainProgram [-verbose] [-ConsensualCFS] [-ModelValidation] [-ModelBuilding] [-WrapperAccuracy] [-WrapperMeanTP] [-Filter] [-SetClassDiscretization cutpointsCommaSeparated] [-boot number] [-fold number] [-repeat number] [-NumClasses number] [-MinInstPerClass number] [-LearningFile filename] [-PerformPrediction filename]");
	        else
	            System.out.println("Success!");
	
		}

 }





